/*
3DHOP - 3D Heritage Online Presenter
Copyright (c) 2014, Marco Callieri - Visual Computing Lab, ISTI - CNR
All rights reserved.    

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Constructs a SphereTrackball object.
 * @class Interactor which implements a full spherical trackball controller.
 */
function SphereTrackball() {
	this.rot = SpiderGL.Math.Quat.identity();
	this.tra = SpiderGL.Math.Vec3.zero();
	this.sca = 1;
	this.center = SpiderGL.Math.Vec3.zero();
	this.radius = 1.0;
	this.lastPoint = SpiderGL.Math.Vec2.zero();
	this.lastTra = this.tra;
	this.lastRot = this.rot;
	this.view = new SglTransformationStack();
}

SphereTrackball.prototype = {

	setup : function (options) {
		options = options || {};
		var opt = sglGetDefaultObject({
			minMaxScale  : [0.00001, 100]
		}, options);
		this._minMaxScale = opt.minMaxScale;
		this._action = SGL_TRACKBALL_NO_ACTION;
		this._matrix = SglMat4.identity();
		this.reset();
	},

    clamp: function(value, low, high) {
      if(value < low) return low;
      if(value > high) return high;
      return value;
    },

    computeMatrix: function() {
		var r = SpiderGL.Math.Quat.to44(this.rot);
		var sr = SpiderGL.Math.Mat4.mul(SpiderGL.Math.Mat4.scaling([this.sca, this.sca, this.sca]), r);
		var s_inv = SpiderGL.Math.Mat4.scaling([1.0 / this.sca, 1.0 / this.sca, 1.0 / this.sca]);
		var c = SpiderGL.Math.Vec3.to4(this.center);
		var temp = SpiderGL.Math.Mat4.mul4(SpiderGL.Math.Mat4.transpose(r), c);
		if  (temp[3] != 0)
		{
			SpiderGL.Math.Vec4.divs(temp, temp[3]);
			temp[3] = 1.0;
		}
		temp = SpiderGL.Math.Mat4.mul4(s_inv, temp);
		if  (temp[3] != 0)
		{
			SpiderGL.Math.Vec4.divs(temp, temp[3]);
			temp[3] = 1.0;
		}		
		var traslV = SpiderGL.Math.Vec3.add(SpiderGL.Math.Vec4.to3(temp), SpiderGL.Math.Vec3.sub(this.tra, this.center));
		var t = SpiderGL.Math.Mat4.translation(traslV);
		this._matrix = SpiderGL.Math.Mat4.mul(sr, t);
    },

	
	_hitViewPlane: function(x, y)
	{
		var temp = SpiderGL.Math.Vec4.zero();
		temp[3] = 1.0;
		var vp = SpiderGL.Math.Mat4.mul4(this.view.modelViewMatrixInverse, temp);
		if  (vp[3] != 0)
		{
			SpiderGL.Math.Vec4.divs(vp, vp[3]);
			vp[3] = 1.0;
		}
		var planeDir = SpiderGL.Math.Vec3.sub(vp, this.center);
		planeDir = SpiderGL.Math.Vec3.normalize(planeDir);
		var planeOffset = SpiderGL.Math.Vec3.dot(planeDir, this.center);
		
		var pp = this.view.unproject([x, y, 0]);
		var lineOrigin = vp;
		var lineDir = SpiderGL.Math.Vec3.sub( SpiderGL.Math.Vec4.to3(pp), vp);
		var k = SpiderGL.Math.Vec3.dot(planeDir, lineDir);
		var r = (planeOffset - SpiderGL.Math.Vec3.dot(planeDir, lineOrigin)) / k;
		return SpiderGL.Math.Vec3.add(lineOrigin, SpiderGL.Math.Vec3.muls(lineDir, r));
	},
	
	
	_intersectionLineSphere: function(lineOrigin, lineDir, viewpoint, hitPoint)
	{
		var newOrigin = SpiderGL.Math.Vec3.sub(lineOrigin, this.center);
		var t1 = lineDir[0] * lineDir[0]; 
		var t2 = lineDir[1] * lineDir[1]; 
		var t3 = lineDir[2] * lineDir[2];
		var t6 = newOrigin[1] * lineDir[1]; 
		var t7 = newOrigin[0] * lineDir[0]; 
		var t8 = newOrigin[2] * lineDir[2];	
		var t15 = this.radius * this.radius;
		var t17 = newOrigin[2] * newOrigin[2]; 
		var t19 = newOrigin[1] * newOrigin[1]; 
		var t21 = newOrigin[0] * newOrigin[0];
		var t28 = 2.0*t7*t6+2.0*t6*t8+2.0*t7*t8+t1*t15-t1*t17-t1*t19-t2*t21+t2*t15-t2*t17-t3*t21+t3*t15-t3*t19;
		if (t28 < 0) return 0;
		var t29 = SpiderGL.Math.sqrt(t28);
		var val0 = 1.0/(t1+t2+t3)*(-t6-t7-t8+t29);
		var val1 = 1.0/(t1+t2+t3)*(-t6-t7-t8-t29);
		var hitPoint1 = SpiderGL.Math.Vec3.add(lineOrigin, SpiderGL.Math.Vec3.muls(lineDir, val0));
		var hitPoint2 = SpiderGL.Math.Vec3.add(lineOrigin, SpiderGL.Math.Vec3.muls(lineDir, val1));
		var dist1 = SpiderGL.Math.Vec3.length(SpiderGL.Math.Vec3.sub(viewpoint, hitPoint1));
		var dist2 = SpiderGL.Math.Vec3.length(SpiderGL.Math.Vec3.sub(viewpoint, hitPoint2));
		if (dist1 < dist2)
		{
			hitPoint[0] = hitPoint1[0];
			hitPoint[1] = hitPoint1[1];
			hitPoint[2] = hitPoint1[2];
		}
		else
		{
			hitPoint[0] = hitPoint2[0];
			hitPoint[1] = hitPoint2[1];
			hitPoint[2] = hitPoint2[2];
		}
		return 2;
	},
	
	
	_intersectionLinePlane: function(planeDir, planeOffset, lineOrigin, lineDir)
	{
		var k = SpiderGL.Math.Vec3.dot(planeDir, lineDir);
		var r = (planeOffset - SpiderGL.Math.Vec3.dot(planeDir, lineOrigin)) / k;
		return SpiderGL.Math.Vec3.add(lineOrigin, SpiderGL.Math.Vec3.muls(lineDir, r));
	},
	
	
	_intersectionHyper: function(viewpoint, planeDir, planeOffset, hitPlane, hitHyper)
	{
		var hitplaney = SpiderGL.Math.Vec3.length(SpiderGL.Math.Vec3.sub(this.center, hitPlane));
		var viewpointx = SpiderGL.Math.Vec3.length(SpiderGL.Math.Vec3.sub(this.center, viewpoint));
		var a = hitplaney / viewpointx;
		var b = -hitplaney;
		var c = this.radius * this.radius / 2.0;
		var delta = b * b - 4.0 * a * c;
		var x1, x2, xval, yval;
		if (delta > 0)
		{
			delta = SpiderGL.Math.sqrt(delta);
			x1 = (-b - delta) / (2.0 * a); 
			x2 = (-b + delta) / (2.0 * a);
			xval = x1;
			yval = c / xval;
		}
		else 
			return 0;
		var dirRadial = SpiderGL.Math.Vec3.normalize(SpiderGL.Math.Vec3.sub(hitPlane, this.center));
		var dirView = SpiderGL.Math.Vec3.normalize(planeDir);
		var offsetx = SpiderGL.Math.Vec3.muls(dirView, xval);
		var offsety = SpiderGL.Math.Vec3.muls(dirRadial, yval);
		var point = SpiderGL.Math.Vec3.add(this.center, SpiderGL.Math.Vec3.add(offsetx, offsety));
		hitHyper[0] = point[0];
		hitHyper[1] = point[1];
		hitHyper[2] = point[2];
		return 1;
	},
	
	_angle: function(p1, p2)
	{
		var w =  SpiderGL.Math.Vec3.length(p1) * SpiderGL.Math.Vec3.length(p2);
		if (w == 0) return -1;
		var t = SpiderGL.Math.Vec3.dot(p1, p2) / w;
		if (t > 1)
			t = 1;
		else if (t < -1)
			t = -1
		return SpiderGL.Math.acos(t);
	},
	
	_hitSphere: function(x, y)
	{
		var temp = SpiderGL.Math.Vec4.zero();
		temp[3] = 1.0;
		var viewpoint = SpiderGL.Math.Mat4.mul4(this.view.modelViewMatrixInverse, temp);
		if  (viewpoint[3] != 0)
		{
			SpiderGL.Math.Vec4.divs(viewpoint, viewpoint[3]);
			viewpoint[3] = 1.0;
		}
		var planeDir = SpiderGL.Math.Vec3.sub(viewpoint, this.center);
		planeDir = SpiderGL.Math.Vec3.normalize(planeDir);
		var planeOffset = SpiderGL.Math.Vec3.dot(planeDir, this.center);
		
		var pp = this.view.unproject([x, y, 0]);
		var lineOrigin = viewpoint;
		var lineDir = SpiderGL.Math.Vec3.sub( SpiderGL.Math.Vec4.to3(pp), viewpoint);
		
		var hitSphere = SpiderGL.Math.Vec3.zero();
		var hitHyper = SpiderGL.Math.Vec3.zero();
		
		var numSphere = this._intersectionLineSphere(lineOrigin, lineDir, viewpoint, hitSphere);
		var hitPlane = this._intersectionLinePlane(planeDir, planeOffset, lineOrigin, lineDir);
		var numHyper = this._intersectionHyper(viewpoint, planeDir, planeOffset, hitPlane, hitHyper);
		
		if (numSphere == 0 && numHyper == 0)
		{
			console.log("problem");			
		}
		if (numSphere > 0 && numHyper == 0)
			return hitSphere;
		if (numHyper > 0 && numSphere == 0)
			return hitHyper; 
		var angle = this._angle(planeDir, SpiderGL.Math.Vec3.sub(hitSphere, this.center));
		if (angle < SpiderGL.Math.PI / 4.0)
			return hitSphere;
		else
			return hitHyper; 
	},
	
	
	getState : function () {
		return this._sphereMatrix;
	},

	setState : function (newstate) {
		this._sphereMatrix = newstate;
		this._computeMatrix();
	},
	
	animateToState : function (newstate) {
		this._sphereMatrix = newstate;
		this._computeMatrix();
	},
	
	setStartingPoint: function(x,y){
		this.lastPoint = [x, y];
		this.lastTra = this.tra;
		this.lastRot = this.rot;
	},
	

	get action()  { return this._action; },

	set action(a) { this._action = a; },

	get matrix() { return this._matrix; },

	reset : function ()
	{
		this.rot = SpiderGL.Math.Quat.identity();
		this.tra = SpiderGL.Math.Vec3.zero();
		this.sca = 1.0;
		this.center = SpiderGL.Math.Vec3.zero();
		this.radius = 1.0;
		this.lastTra = this.tra;
		this.lastRot = this.rot;
		
		this._matrix = SglMat4.identity();
		this._action = SGL_TRACKBALL_NO_ACTION;
		this._new_action = true;
		this.computeMatrix();
	},

	track : function(m, x, y, z) 
	{
		switch (this._action) {
			case SGL_TRACKBALL_ROTATE:
				this.rotate(x, y);
			break;

			case SGL_TRACKBALL_PAN:
				this.pan(x, y);
			break;

			case SGL_TRACKBALL_DOLLY:
				this.dolly(z);
			break;

			case SGL_TRACKBALL_SCALE:
				this.scale(z);
			break;

			default:
			break;
		}
		this.computeMatrix();
	},

	rotate : function(x, y) 
	{
		var hitOld = this._hitSphere(this.lastPoint[0], this.lastPoint[1]);
		var hitNew = this._hitSphere(x, y);
		var offsetOld = SpiderGL.Math.Vec3.sub(hitOld, this.center);
		var offsetNew = SpiderGL.Math.Vec3.sub(hitNew, this.center);
		var axis = SpiderGL.Math.Vec3.cross(offsetNew, offsetOld);
		axis = SpiderGL.Math.Vec3.normalize(axis);
		var distance = SpiderGL.Math.Vec3.length(SpiderGL.Math.Vec3.sub(hitNew, hitOld))/this.radius;
		var phi = this._angle(offsetNew, offsetOld);
		var quater = SpiderGL.Math.Quat.fromAngleAxis(-phi, axis);
		this.rot = SpiderGL.Math.Quat.mul(quater, this.lastRot);
	},

	pan : function(x, y) 
	{
		var hitOld = this._hitViewPlane(this.lastPoint[0], this.lastPoint[1]);
		var hitNew = this._hitViewPlane(x, y);
		this.translate(SpiderGL.Math.Vec3.sub(hitNew, hitOld));
	},

	dolly : function(z)
	{
		var temp = SpiderGL.Math.Vec4.zero();
		temp[3] = 1.0;
		var vp = SpiderGL.Math.Mat4.mul4(this.view.modelViewMatrixInverse, temp);
		if  (vp[3] != 0)
		{
			SpiderGL.Math.Vec4.divs(vp, vp[3]);
			vp[3] = 1.0;
		}
		var planeDir = SpiderGL.Math.Vec3.sub(vp, this.center);
		planeDir = SpiderGL.Math.Vec3.normalize(planeDir);
		this.translate(SpiderGL.Math.Vec3.muls(planeDir, 0.1 * z));
	},

	recentering: function(x, s)
	{
		this.translate(x);
		var newScale = this.sca * s;
		this.sca = this.clamp(newScale, this._minMaxScale[0], this._minMaxScale[1]);
		this.computeMatrix();
	},
	
	scale : function(s)
	{
		var newScale = this.sca * sglPow(1.2, s);
		this.sca = this.clamp(newScale, this._minMaxScale[0], this._minMaxScale[1]);
	},
	
	translate : function(x)
	{
		var q = SpiderGL.Math.Quat.dup(this.rot);
		q = SpiderGL.Math.Quat.inverse(q);
		var tmp = [x[0], x[1], x[2], 0];
		tmp = SpiderGL.Math.Quat.mul(q, SpiderGL.Math.Quat.mul(tmp, this.rot));
		var newPoint = [tmp[0], tmp[1], tmp[2]];
		this.tra = SpiderGL.Math.Vec3.add(this.lastTra, SpiderGL.Math.Vec3.divs(newPoint, this.sca));		
	}
};
